{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "application/javascript": [
       "IPython.notebook.set_autosave_interval(60000)"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Autosaving every 60 seconds\n"
     ]
    }
   ],
   "source": [
    "%autosave 60\n",
    "import numpy as np\n",
    "from threading import Thread\n",
    "from functools import partial\n",
    "import multiprocessing\n",
    "import multiprocessing.pool"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "N_CPU = multiprocessing.cpu_count()\n",
    "REPEAT = 16 * N_CPU"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "world\n",
      "hello\n",
      "!\n"
     ]
    }
   ],
   "source": [
    "from time import sleep\n",
    "\n",
    "def f():\n",
    "    sleep(3)\n",
    "    print('hello')\n",
    "    \n",
    "t = Thread(target=f)\n",
    "t.start()\n",
    "print('world')\n",
    "t.join()\n",
    "print('!')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "35.7 ms ± 597 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)\n"
     ]
    }
   ],
   "source": [
    "A = np.random.normal(0, 1, (64, 64, 5, 5))\n",
    "\n",
    "%timeit np.linalg.eigvals(A)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "690 ms ± 34.2 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit A = np.random.normal(0, 1, (64, 64, 5, 5))\n",
    "with multiprocessing.pool.ThreadPool() as pool:\n",
    "    pool.map(lambda _: np.linalg.eigvals(A), range(REPEAT))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Eigen:\n",
    "    def __init__(self):\n",
    "        self.A = np.random.normal(0, 1, (64, 64, 5, 5))\n",
    "    \n",
    "    def __call__(self, x):\n",
    "        A = self.A.copy()\n",
    "        A[0, 0] += x\n",
    "        vals = np.linalg.eigvals(A)\n",
    "        return np.linalg.norm(vals)\n",
    "\n",
    "f = Eigen()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1.42 s ± 64.3 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit\n",
    "with multiprocessing.pool.ThreadPool() as pool:\n",
    "    pool.map(f, range(REPEAT))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "4.75 s ± 143 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%timeit [f(x) for x in range(REPEAT)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "def golden_ratio(n):\n",
    "    s = 1.0\n",
    "    for _ in range(n):\n",
    "        s = 1.0 + 1.0 / s\n",
    "    return s"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "882 ms ± 32.7 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%timeit [golden_ratio(100_000) for _ in range(REPEAT)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "930 ms ± 31.5 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit\n",
    "with multiprocessing.pool.ThreadPool() as pool:\n",
    "    pool.map(lambda _: golden_ratio(100_000), range(REPEAT))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [],
   "source": [
    "# GIL"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "# GIL отпущен во время всех I/O операций\n",
    "\n",
    "# with open()...\n",
    "\n",
    "# GIL отпущен в некоторых функциях некоторых \"нативных\" модулей\n",
    "# типа numpy, scipy, ..."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "262 ms ± 2.34 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit\n",
    "with multiprocessing.Pool() as pool:\n",
    "    pool.map(golden_ratio, [100_000] * REPEAT)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1.49 s ± 6.08 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit a = np.random.rand(10_000_000)\n",
    "for _ in range(REPEAT):\n",
    "    np.sum(a)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "8.37 s ± 104 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)\n"
     ]
    }
   ],
   "source": [
    "%%timeit a = np.random.rand(10_000_000)\n",
    "with multiprocessing.Pool() as pool:\n",
    "    pool.map(np.sum, [a] * REPEAT)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [],
   "source": [
    "import pickle"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 28,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "b'\\x80\\x03K{.'"
      ]
     },
     "execution_count": 28,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "x = 123\n",
    "pickle.dumps(x)\n",
    "# pickle.dumps(lambda x: x**2) - ошибка"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 29,
   "metadata": {},
   "outputs": [],
   "source": [
    "from functools import partial"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 30,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[array([0.89921603]), array([0.89921603]), array([0.89921603]), array([0.89921603]), array([0.89921603]), array([0.89921603]), array([0.89921603]), array([0.89921603])]\n"
     ]
    }
   ],
   "source": [
    "with multiprocessing.Pool() as pool:\n",
    "    result = pool.map(np.random.rand, [1] * N_CPU)\n",
    "print(result)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 31,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[0.8992160348470516,\n",
       " 0.9551956229774841,\n",
       " 0.07824628133895539,\n",
       " 0.823248685458748,\n",
       " 0.13479980996708463,\n",
       " 0.3900273918893963,\n",
       " 0.07412678098979342,\n",
       " 0.8652795249825378]"
      ]
     },
     "execution_count": 31,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "[np.random.rand() for _ in range(N_CPU)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 33,
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[0]\n",
      "[4]\n",
      "[8]\n",
      "[12]\n",
      "[8, 9]\n",
      "[24]\n",
      "[16]\n",
      "[20]\n",
      "[28]\n",
      "[0, 1]\n",
      "[4, 5]\n",
      "[12, 13]\n",
      "[20, 21]\n",
      "[12, 13, 14]\n",
      "[8, 9, 10]\n",
      "[4, 5, 6]\n",
      "[24, 25]\n",
      "[16, 17]\n",
      "[20, 21, 22]\n",
      "[4, 5, 6, 7]\n",
      "[12, 13, 14, 15]\n",
      "[24, 25, 26]\n",
      "[8, 9, 10, 11]\n",
      "[0, 1, 2]\n",
      "[16, 17, 18]\n",
      "[20, 21, 22, 23]\n",
      "[16, 17, 18, 19]\n",
      "[4, 5, 6, 7, 36]\n",
      "[24, 25, 26, 27]\n",
      "[0, 1, 2, 3]\n",
      "[12, 13, 14, 15, 32]\n",
      "[16, 17, 18, 19, 48]\n",
      "[12, 13, 14, 15, 32, 33]\n",
      "[20, 21, 22, 23, 44]\n",
      "[16, 17, 18, 19, 48, 49]\n",
      "[12, 13, 14, 15, 32, 33, 34]\n",
      "[8, 9, 10, 11, 40]\n",
      "[24, 25, 26, 27, 56]\n",
      "[4, 5, 6, 7, 36, 37]\n",
      "[0, 1, 2, 3, 52]\n",
      "[28, 29, 30]\n",
      "[24, 25, 26, 27, 56, 57]\n",
      "[20, 21, 22, 23, 44, 45]\n",
      "[4, 5, 6, 7, 36, 37, 38]\n",
      "[28, 29]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61]\n",
      "[16, 17, 18, 19, 48, 49, 50]\n",
      "[8, 9, 10, 11, 40, 41]\n",
      "[20, 21, 22, 23, 44, 45, 46]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35]\n",
      "[0, 1, 2, 3, 52, 53]\n",
      "[24, 25, 26, 27, 56, 57, 58]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47]\n",
      "[8, 9, 10, 11, 40, 41, 42]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68]\n",
      "[28, 29, 30, 31]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73]\n",
      "[28, 29, 30, 31, 76]\n",
      "[0, 1, 2, 3, 52, 53, 54]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74]\n",
      "[28, 29, 30, 31, 76, 77]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81]\n",
      "[0, 1, 2, 3, 52, 53, 54, 55]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85]\n",
      "[28, 29, 30, 31, 76, 77, 78]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82, 83]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74, 75]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82, 83, 96]\n",
      "[28, 29, 30, 31, 76, 77, 78, 79]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70, 71]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86, 87]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82, 83, 96, 97]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70, 71, 108]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74, 75, 104]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101]\n",
      "[28, 29, 30, 31, 76, 77, 78, 79, 112]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70, 71, 108, 109]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74, 75, 104, 105]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102]\n",
      "[0, 1, 2, 3, 52, 53, 54, 55, 88]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82, 83, 96, 97, 98]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86, 87, 116]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74, 75, 104, 105, 106]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70, 71, 108, 109, 110]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102, 103]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94, 95]\n",
      "[0, 1, 2, 3, 52, 53, 54, 55, 88, 89]\n",
      "[20, 21, 22, 23, 44, 45, 46, 47, 68, 69, 70, 71, 108, 109, 110, 111]\n",
      "[16, 17, 18, 19, 48, 49, 50, 51, 72, 73, 74, 75, 104, 105, 106, 107]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86, 87, 116, 117]\n",
      "[8, 9, 10, 11, 40, 41, 42, 43, 80, 81, 82, 83, 96, 97, 98, 99]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94, 95, 124]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86, 87, 116, 117, 118]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94, 95, 124, 125]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102, 103, 120]\n",
      "[0, 1, 2, 3, 52, 53, 54, 55, 88, 89, 90]\n",
      "[24, 25, 26, 27, 56, 57, 58, 59, 84, 85, 86, 87, 116, 117, 118, 119]\n",
      "[0, 1, 2, 3, 52, 53, 54, 55, 88, 89, 90, 91]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94, 95, 124, 125, 126]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102, 103, 120, 121]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102, 103, 120, 121, 122]\n",
      "[4, 5, 6, 7, 36, 37, 38, 39, 60, 61, 62, 63, 92, 93, 94, 95, 124, 125, 126, 127]\n",
      "[12, 13, 14, 15, 32, 33, 34, 35, 64, 65, 66, 67, 100, 101, 102, 103, 120, 121, 122, 123]\n",
      "[28, 29, 30, 31, 76, 77, 78, 79, 112, 113]\n",
      "[28, 29, 30, 31, 76, 77, 78, 79, 112, 113, 114]\n",
      "[28, 29, 30, 31, 76, 77, 78, 79, 112, 113, 114, 115]\n",
      "[]\n"
     ]
    }
   ],
   "source": [
    "a = []\n",
    "\n",
    "def f(x):\n",
    "    a.append(x)\n",
    "    print(a)\n",
    "    \n",
    "\n",
    "with multiprocessing.Pool() as pool:\n",
    "    pool.map(f, range(REPEAT))\n",
    "    \n",
    "print(a)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 34,
   "metadata": {},
   "outputs": [],
   "source": [
    "# cloudpickle\n",
    "# joblib\n",
    "# dask"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
